/*
 * An AVL binary search tree, the nodes are allocated in large arrays.
 * Using bits and pieces of code from avl-2.0.2a/avl.c .
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "pqt.h"

#define PQT_BLOCKSIZE 1024
#define PQT_NUMBEROFBLOCKS 128

#define PQT_AVL_DEBUG 0

#if PQT_AVL_DEBUG
   #define PQT_AVL_ASSERT(v,s) if (v) printf (s)
   #define PQT_AVL_ASSERTQ(v,s) if (v) { printf (s); return; }
#else
   #define PQT_AVL_ASSERT(v,s)
   #define PQT_AVL_ASSERTQ(v,s)
#endif

typedef struct PQT__tmp__node {
   PQT_pixel pixel;               /* pixel data */
   struct PQT__tmp__node *left;   /* left child node */
   struct PQT__tmp__node *right;  /* right child node */
   signed char avl_balance;       /* balance factor */
} PQT__node, *PQT_node;

struct PQT__tmp__priorityqueue {
   PQT_node root;           /* pointer to the root */
   PQT_node freelist;       /* pointer to the list of free nodes; node->left points to the next empty node in the list */
   int nelements;           /* number of elements on the queue */
   int blocksize;           /* number of nodes allocated in each block */
   int nblocks;             /* number of blocks allocated */
   int arraysize;           /* number of elements in pointer array */
   PQT_node *blocks;        /* array of pointers to the blocks */
};


static PQT_error PQT__NewBlock
(
   PQT_priorityqueue tree
)
{
   PQT_node block;
   int ii;

   /* Check & extend pointer array */
   if (tree->nblocks >= tree->arraysize) {
      PQT_node* tmpblocks = 0;
      tmpblocks = (PQT_node*)realloc (tree->blocks, 2 * tree->arraysize * sizeof (PQT_node));
      if (!tmpblocks) {
         return PQT_PROBLEM;
      }
      tree->blocks = tmpblocks;
      tree->arraysize *= 2;
   }

   /* Allocate & initialise tree block */
   block = (PQT_node)malloc (tree->blocksize * sizeof (PQT__node));
   if (!block) {
      return PQT_PROBLEM;
   }
   tree->blocks[tree->nblocks] = block;
   tree->nblocks++;
   for (ii=0; ii<tree->blocksize; ii++) {
      block[ii].left = &(block[ii+1]);
   }
   block[tree->blocksize-1].left = tree->freelist;
   tree->freelist = block;

   return PQT_OK;
}


static PQT_error PQT__NewNode
(
   PQT_node *node,
   PQT_priorityqueue tree,
   void* pointer,
   float value
)
{
   PQT_node tmpnode;

   if (!tree->freelist) {
      PQT_error retval = PQT__NewBlock (tree);
      if (retval != PQT_OK) {
         return retval;
      }
   }

   tmpnode = tree->freelist;
   tree->freelist = tree->freelist->left;
   tmpnode->left  = 0;
   tmpnode->right = 0;
   tmpnode->avl_balance = 0;
   tmpnode->pixel.pointer = pointer;
   tmpnode->pixel.value   = value;
   *node = tmpnode;

   return PQT_OK;
}


static void PQT__FreeNode
(
   PQT_node node,
   PQT_priorityqueue tree
)
{
   node->left = tree->freelist;
   tree->freelist = node;
}


PQT_error PQT_New
(
   PQT_priorityqueue *tree,
   int blocksize
)
{
   PQT_priorityqueue tmptree = 0;

   if (blocksize <= 0) {
      blocksize = PQT_BLOCKSIZE;
   }

   tmptree = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmptree) goto PQT_ERROR;
   tmptree->nelements = 0;
   tmptree->root = 0;
   tmptree->blocksize = blocksize;
   tmptree->nblocks = 0;
   tmptree->freelist = 0;
   tmptree->blocks = (PQT_node*)malloc (PQT_NUMBEROFBLOCKS * sizeof (PQT_node*));
   if (!tmptree->blocks) goto PQT_ERROR;
   tmptree->arraysize = PQT_NUMBEROFBLOCKS;

   *tree = tmptree;
   tmptree = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmptree) {
      if (tmptree->blocks) {
         free (tmptree->blocks);
      }
      free (tmptree);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue tree
)
{
   int ii;
   if (tree) {
      if (tree->blocks) {
         for (ii=0; ii<tree->nblocks; ii++) {
            free (tree->blocks[ii]);
         }
         free (tree->blocks);
      }
      free (tree);
   }
}


#if PQT_AVL_DEBUG

static PQT_boolean PQT__CountNodes
(
   PQT_node node,
   int *n,
   int *d
)
{
   int nl, nr;
   int dl, dr;
   PQT_boolean error = PQT_FALSE;

   if (node->left) {
      if (PQT__CountNodes (node->left, &nl, &dl)) {
         error = PQT_TRUE;
      }
   }
   else {
      nl = dl = 0;
   }
   if (node->right) {
      if (PQT__CountNodes (node->right, &nr, &dr)) {
         error = PQT_TRUE;
      }
   }
   else {
      nr = dr = 0;
   }

   if (node->avl_balance != dr-dl) {
      printf ("ASSERTION FAILED!!! (#02) [balance number for node is wrong: %d != %d-%d]\n", node->avl_balance, dr, dl);
      error = PQT_TRUE;
   }

   *n = 1 + nl + nr;
   *d = 1 + (dl>dr ? dl : dr);

   return error;
}


static void PQT__CheckTree
(
   PQT_priorityqueue tree
)
{
   int n, d;
   PQT_boolean error;
   if (tree->root) {
      error = PQT__CountNodes (tree->root, &n, &d);
      if (n != tree->nelements) {
         printf ("ASSERTION FAILED!!! (#01) [%d items found, %d presumed]\n", n, tree->nelements);
         error = PQT_TRUE;
      }
   }
}

#endif /* PQT_AVL_DEBUG */


static void PQT__RebalanceLeft
(
   PQT_node *node
)
{
   /* Rebalance left-heavy node */
   PQT_node y, x, w;
   y = *node;
   PQT_AVL_ASSERTQ ( y->avl_balance != -2, "ASSERTION FAILED!!! (#03)\n" );
   x = y->left;
   PQT_AVL_ASSERTQ ( !x, "ASSERTION FAILED!!! (#04)\n" );
   switch (x->avl_balance) {
      case 1:
         w = x->right;
         PQT_AVL_ASSERTQ ( !w, "ASSERTION FAILED!!! (#05)\n" );
         x->right = w->left;
         w->left = x;
         y->left = w->right;
         w->right = y;
         (*node) = w;
         x->avl_balance = 0;
         y->avl_balance = 0;
         if (w->avl_balance == -1)
            y->avl_balance = 1;
         else if (w->avl_balance == 1)
            x->avl_balance = -1;
         w->avl_balance = 0;
         break;
      case 0:
      case -1:
         y->left = x->right;
         x->right = y;
         (*node) = x;
         if (x->avl_balance == 0) {
            x->avl_balance = 1;
            y->avl_balance = -1;
         }
         else {
            x->avl_balance = 0;
            y->avl_balance = 0;
         }
         break;
   }
}


static void PQT__RebalanceRight
(
   PQT_node *node
)
{
   /* Rebalance right-heavy node */
   PQT_node y, x, w;
   y = *node;
   PQT_AVL_ASSERTQ ( y->avl_balance != 2, "ASSERTION FAILED!!! (#06)\n" );
   x = y->right;
   PQT_AVL_ASSERTQ ( !x, "ASSERTION FAILED!!! (#07)\n" );
   switch (x->avl_balance) {
      case -1:
         w = x->left;
         PQT_AVL_ASSERTQ ( !w, "ASSERTION FAILED!!! (#08)\n" );
         x->left = w->right;
         w->right = x;
         y->right = w->left;
         w->left = y;
         (*node) = w;
         x->avl_balance = 0;
         y->avl_balance = 0;
         if (w->avl_balance == 1)
            y->avl_balance = -1;
         else if (w->avl_balance == -1)
            x->avl_balance = 1;
         w->avl_balance = 0;
         break;
      case 0:
      case 1:
         y->right = x->left;
         x->left = y;
         (*node) = x;
         if (x->avl_balance == 0) {
            x->avl_balance = -1;
            y->avl_balance = 1;
         }
         else {
            x->avl_balance = 0;
            y->avl_balance = 0;
         }
         break;
   }
}


static PQT_boolean PQT__Insert
(
   PQT_node *node,
   PQT_node newnode
)
{
   PQT_boolean heightchange;

   /* Equal values inserted to the right -- as a minimum-pop queue this is fifo for equal-valued items */
   if ((*node)->pixel.value <= newnode->pixel.value) {
      /* to the right child */
      if ((*node)->right) {
         heightchange = PQT__Insert (&((*node)->right), newnode);
         if (heightchange) {
            (*node)->avl_balance += 1;
            if ((*node)->avl_balance > 1) {
               PQT__RebalanceRight (node);
               heightchange = PQT_FALSE;
            }
            if ((*node)->avl_balance == 0) {
               heightchange = PQT_FALSE;
            }
         }
      }
      else {
         (*node)->right = newnode;
         (*node)->avl_balance += 1;
         PQT_AVL_ASSERT ( (*node)->avl_balance > 1, "ASSERTION FAILED!!! (#09)\n" );
         heightchange = ((*node)->avl_balance != 0);
      }
   }
   else {
      /* to the left child */
      if ((*node)->left) {
         heightchange = PQT__Insert (&((*node)->left), newnode);
         if (heightchange) {
            (*node)->avl_balance -= 1;
            if ((*node)->avl_balance < -1) {
               PQT__RebalanceLeft (node);
               heightchange = PQT_FALSE;
            }
            if ((*node)->avl_balance == 0) {
               heightchange = PQT_FALSE;
            }
         }
      }
      else {
         (*node)->left = newnode;
         (*node)->avl_balance -= 1;
         PQT_AVL_ASSERT ( (*node)->avl_balance < -1, "ASSERTION FAILED!!! (#10)\n" );
         heightchange = ((*node)->avl_balance != 0);
      }
   }

   return heightchange;
}


PQT_error PQT_Push
(
   PQT_priorityqueue tree,
   void* pointer,
   float value
)
{
   PQT_node newnode;
   PQT_error retval;

   retval = PQT__NewNode (&newnode, tree, pointer, value);
   if (retval != PQT_OK) {
      return retval;
   }

   if (tree->root) {
      PQT__Insert (&(tree->root), newnode);
   }
   else {
      tree->root = newnode;
   }
   tree->nelements++;

#if PQT_AVL_DEBUG
   PQT__CheckTree (tree);
#endif

   return PQT_OK;
}


static PQT_boolean PQT__DeleteMin
(
   PQT_node *node,
   PQT_priorityqueue tree,
   PQT_pixel *pixel
)
{
   PQT_boolean heightchange = PQT_FALSE;

   if ((*node)->left) {
      if (PQT__DeleteMin (&((*node)->left), tree, pixel)) {
         if (*node) {
            (*node)->avl_balance += 1;
            PQT_AVL_ASSERT ( (*node)->avl_balance > 2, "ASSERTION FAILED!!! (#11)\n" );
            if ((*node)->avl_balance == 2) {
               PQT__RebalanceRight (node);
            }
            heightchange = ((*node)->avl_balance == 0);
         }
      }
   }
   else {
      PQT_node x = (*node)->right;
      *pixel = (*node)->pixel;   /* copies data, not a pointer */
      PQT__FreeNode (*node, tree);
      *node = x;
      heightchange = PQT_TRUE;
   }

   return heightchange;
}


PQT_error PQT_Pop
(
   PQT_priorityqueue tree,
   void* *pointer,
   float *value
)
{
   PQT_pixel pixel;

   if (!tree->root) {
      return PQT_PROBLEM;
   }

   PQT__DeleteMin (&(tree->root), tree, &pixel);
   tree->nelements--;
   if (pointer) {
      *pointer = pixel.pointer;
   }
   if (value) {
      *value = pixel.value;
   }

#if PQT_AVL_DEBUG
   PQT__CheckTree (tree);
#endif

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue tree
)
{
   if (!tree->root) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
