/*
 * A simple binary search tree.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include "pqt.h"

typedef struct PQT__tmp__node {
   PQT_pixel pixel;               /* pixel data */
   struct PQT__tmp__node *left;   /* left child node */
   struct PQT__tmp__node *right;  /* right child node */
} PQT__node, *PQT_node;

struct PQT__tmp__priorityqueue {
   PQT_node root;           /* pointer to the root */
   int nelements;           /* number of elements on the queue */
};


static PQT_error PQT__NewNode
(
   PQT_node *node,
   void* pointer,
   float value
)
{
   PQT_node tmpnode = 0;
   tmpnode = (PQT_node)malloc (sizeof (PQT__node));
   if (!tmpnode) return PQT_PROBLEM;
   tmpnode->left  = 0;
   tmpnode->right = 0;
   tmpnode->pixel.pointer = pointer;
   tmpnode->pixel.value   = value;
   *node = tmpnode;
   return PQT_OK;
}


static void PQT__FreeNode
(
   PQT_node node
)
{
   free (node);
}


static void PQT__RecursiveFreeNode
(
   PQT_node node
)
{
   if (node->left) {
      PQT__RecursiveFreeNode (node->left);
   }
   if (node->right) {
      PQT__RecursiveFreeNode (node->right);
   }
   PQT__FreeNode (node);
}


PQT_error PQT_New
(
   PQT_priorityqueue *tree,
   int blocksize /* ignored */
)
{
   PQT_priorityqueue tmptree = 0;
   tmptree = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmptree) return PQT_PROBLEM;
   tmptree->nelements = 0;
   tmptree->root = 0;
   *tree = tmptree;
   tmptree = 0;
   return PQT_OK;
}


void PQT_Free
(
   PQT_priorityqueue tree
)
{
   if (tree) {
      if (tree->root) {
         PQT__RecursiveFreeNode (tree->root);
      }
      free (tree);
   }
}


static void PQT__Insert
(
   PQT_node node,
   PQT_node newnode
)
{
   /* Equal values inserted to the right -- as a minimum-pop queue this is fifo for equal-valued items */
   if (node->pixel.value <= newnode->pixel.value) {
      /* to the right child */
      if (node->right) {
         PQT__Insert (node->right, newnode);
      }
      else {
         node->right = newnode;
      }
   }
   else {
      /* to the left child */
      if (node->left) {
         PQT__Insert (node->left, newnode);
      }
      else {
         node->left = newnode;
      }
   }
}


PQT_error PQT_Push
(
   PQT_priorityqueue tree,
   void* pointer,
   float value
)
{
   PQT_node newnode;
   PQT_error retval;

   retval = PQT__NewNode (&newnode, pointer, value);
   if (retval != PQT_OK) {
      return retval;
   }

   if (tree->root) {
      PQT__Insert (tree->root, newnode);
   }
   else {
      tree->root = newnode;
   }
   tree->nelements++;

   return PQT_OK;
}


static void PQT__DeleteMin
(
   PQT_node *node,
   PQT_pixel *pixel
)
{
   if ((*node)->left) {
      PQT__DeleteMin (&((*node)->left), pixel);
   }
   else {
      PQT_node tmp = 0;
      *pixel = (*node)->pixel;         /* copies data, not a pointer */
      tmp = (*node)->right;
      PQT__FreeNode (*node);
      *node = tmp;
   }
}


PQT_error PQT_Pop
(
   PQT_priorityqueue tree,
   void* *pointer,
   float *value
)
{
   PQT_pixel pixel;

   if (!tree->root) {
      return PQT_PROBLEM;
   }

   PQT__DeleteMin (&(tree->root), &pixel);
   tree->nelements--;
   if (pointer) {
      *pointer = pixel.pointer;
   }
   if (value) {
      *value = pixel.value;
   }

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue tree
)
{
   if (!tree->root) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
