/*
 * A heap in an array that is dynamically resized to fit the data.
 *
 * Change "PQT_OFFSET" to some positive value to try out LaMarca and
 * Ladner's (1996) idea for reducing cache misses.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include "pqt.h"

#define PQT_BLOCKSIZE 1024

#define PQT_OFFSET 0

struct PQT__tmp__priorityqueue {
   PQT_pixel *pixels;       /* array with info for each pixel */
   int arraysize;           /* number of elements allocated */
   int nelements;           /* number of elements on the queue */
};

PQT_error PQT_New
(
   PQT_priorityqueue *heap,
   int blocksize
)
{
   PQT_priorityqueue tmpheap = 0;
   PQT_pixel* tmppixels = 0;

   if (blocksize <= 0) {
      blocksize = PQT_BLOCKSIZE;
   }

   /* Allocate & initialise structure */
   tmpheap = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmpheap) goto PQT_ERROR;
   tmpheap->pixels = 0;
   tmpheap->nelements = 0;
   tmpheap->arraysize = 0;

   /* Allocate & initialise heap */
   tmppixels = (PQT_pixel*)malloc ((blocksize+PQT_OFFSET) * sizeof (PQT_pixel));
   if (!tmppixels) goto PQT_ERROR;
   tmpheap->pixels = tmppixels+PQT_OFFSET;
   tmppixels = 0;
   tmpheap->arraysize = blocksize;

   /* Do the switch */
   *heap = tmpheap;
   tmpheap = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpheap) {
      if (tmppixels) {
         free (tmppixels);
      }
      free (tmpheap);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue heap
)
{
   if (heap) {
      free (heap->pixels - PQT_OFFSET);
      free (heap);
   }
}


static PQT_error PQT__Enlarge
(
   PQT_priorityqueue heap
)
{
   PQT_pixel* tmppixels = 0;

   /* Allocate & initialise heap block */
   tmppixels = (PQT_pixel*)realloc (heap->pixels - PQT_OFFSET,
                                    (2 * heap->arraysize + PQT_OFFSET) * sizeof (PQT_pixel));
   if (!tmppixels) return PQT_PROBLEM;
   heap->pixels = tmppixels + PQT_OFFSET;
   heap->arraysize *= 2;
   tmppixels = 0;

   return PQT_OK;
}


/* This one assumes item p is where the data is at, and moves this item
 * up (towards the root) to sort it */
static void PQT__MoveUp
(
   PQT_priorityqueue heap,
   int p,                    /* item to be sorted */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2;       /* for parent */

   while (p > 0) {
      p2 = (p-1)/2; /* parent */
      if (value < heap->pixels[p2].value) {
         heap->pixels[p].pointer = heap->pixels[p2].pointer;
         heap->pixels[p].value   = heap->pixels[p2].value;
         p = p2;
      }
      else {
         /* we're done */
         break;
      }
   }
   heap->pixels[p].pointer = pointer;
   heap->pixels[p].value   = value;
}

/* This one assumes item p is where the data is at, and moves this item
 * down (away from the root) to sort it */
static void PQT__MoveDown
(
   PQT_priorityqueue heap,
   int p,                    /* item to be sorted */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2 = 2*p+1;   /* left child. right = 2*p+2 = p2+1 */
   while (p2 < heap->nelements) {
      if (p2+1 < heap->nelements) {
         if (heap->pixels[p2+1].value < heap->pixels[p2].value) {
            p2 = p2+1;
         }
      }
      /* compare p to p2 */
      if (heap->pixels[p2].value < value) {
         heap->pixels[p].pointer = heap->pixels[p2].pointer;
         heap->pixels[p].value   = heap->pixels[p2].value;
         p = p2;
         p2 = 2*p+1;
      }
      else {
         /* we're done */
         break;
      }
   }
   heap->pixels[p].pointer = pointer;
   heap->pixels[p].value = value;
}


PQT_error PQT_Push
(
   PQT_priorityqueue heap,
   void* pointer,
   float value
)
{
   int p;
   PQT_error retval;

   if (heap->nelements >= heap->arraysize) {
      retval = PQT__Enlarge (heap);
      if (retval != PQT_OK) {
         return retval;
      }
   }
   p = heap->nelements;       /* position of insertion */
   heap->nelements++;
   PQT__MoveUp (heap, p, pointer, value);

   return PQT_OK;
}


PQT_error PQT_Pop
(
   PQT_priorityqueue heap,
   void* *pointer,
   float *value
)
{
   int p;

   if (heap->nelements == 0) {
      return PQT_PROBLEM;
   }
   if (pointer) {
      *pointer = heap->pixels[0].pointer;
   }
   if (value) {
      *value = heap->pixels[0].value;
   }
   heap->nelements--;
   p = heap->nelements;       /* position of deletion */
   PQT__MoveDown (heap, 0, heap->pixels[p].pointer, heap->pixels[p].value);

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue heap
)
{
   if (heap->nelements == 0) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
