/*
 * A heap distributed over blocks, with a table of pointers to the blocks.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include "pqt.h"

#define PQT_BLOCKSIZE 1024
#define PQT_NUMBEROFBLOCKS 128

typedef struct PQT__tmp__PixelHeapBlock {
   PQT_pixel *pixels;  /* array with info for each pixel */
} PQT__block, *PQT_block;

struct PQT__tmp__priorityqueue {
   PQT_block *blocks;
   int nelements;           /* number of elements on the queue */
   int blocksize;           /* number of elements allocated in each block */
   int nblocks;             /* number of blocks allocated */
   int arraysize;           /* number of elements in pointer array */
};


PQT_error PQT_New
(
   PQT_priorityqueue *heap,
   int blocksize
)
{
   PQT_priorityqueue tmpheap = 0;
   PQT_block tmpblock = 0;

   if (blocksize <= 0) {
      blocksize = PQT_BLOCKSIZE;
   }

   /* Allocate & initialise structure */
   tmpheap = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmpheap) goto PQT_ERROR;
   tmpheap->blocksize = blocksize;
   tmpheap->nelements = 0;
   tmpheap->nblocks = 0;
   tmpheap->blocks = (PQT_block*)malloc (PQT_NUMBEROFBLOCKS * sizeof (PQT_block));
   if (!tmpheap->blocks) goto PQT_ERROR;
   tmpheap->arraysize = PQT_NUMBEROFBLOCKS;

   /* Allocate & initialise first heap block */
   tmpblock = (PQT_block)malloc (sizeof (PQT__block));
   if (!tmpblock) goto PQT_ERROR;
   tmpblock->pixels = 0;
   tmpblock->pixels = (PQT_pixel*)malloc (tmpheap->blocksize * sizeof (PQT_pixel));
   if (!tmpblock->pixels) goto PQT_ERROR;
   tmpheap->blocks[0] = tmpblock;
   tmpheap->nblocks = 1;

   /* Do the switch */
   *heap = tmpheap;
   tmpheap = 0;
   tmpblock = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpblock) {
      if (tmpblock->pixels) {
         free (tmpblock->pixels);
      }
      free (tmpblock);
   }
   if (tmpheap) {
      if (tmpheap->blocks) {
         free (tmpheap->blocks);
      }
      free (tmpheap);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue heap
)
{
   int ii;

   if (heap) {
      for (ii=0; ii<heap->nblocks; ii++) {
         free (heap->blocks[ii]->pixels);
         free (heap->blocks[ii]);
      }
      free (heap->blocks);
      free (heap);
   }
}


static PQT_error PQT__Enlarge
(
   PQT_priorityqueue heap
)
{
   PQT_block tmpblock = 0;

   /* Check & extend pointer array */
   if (heap->nblocks >= heap->arraysize) {
      PQT_block* tmpblocks = 0;
      tmpblocks = (PQT_block*)realloc (heap->blocks, 2 * heap->arraysize * sizeof (PQT_block));
      if (!tmpblocks) {
         return PQT_PROBLEM;
      }
      heap->blocks = tmpblocks;
      heap->arraysize *= 2;
   }

   /* Allocate & initialise heap block */
   tmpblock = (PQT_block)malloc (sizeof (PQT__block));
   if (!tmpblock) goto PQT_ERROR;
   tmpblock->pixels = 0;
   tmpblock->pixels = (PQT_pixel*)malloc (heap->blocksize * sizeof (PQT_pixel));
   if (!tmpblock->pixels) goto PQT_ERROR;

   /* Set pointers */
   heap->blocks[heap->nblocks] = tmpblock;
   heap->nblocks++;
   tmpblock = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpblock) {
      if (tmpblock->pixels) {
         free (tmpblock->pixels);
      }
      free (tmpblock);
   }
   return PQT_PROBLEM;
}


/* This one assumes item p is where the data is at, and moves this item
 * up (towards the root) to sort it */
static void PQT__MoveUp
(
   PQT_priorityqueue heap,
   int p,                    /* item to be sorted */
   PQT_block block,          /* it's in this block */
   int b,                    /* this is the block number */
   int i,                    /* this is the index into the block */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2, b2, i2;       /* for parent */
   PQT_block block2;

   while (p > 0) {
      p2 = (p-1)/2; /* parent */
      b2 = p2 / heap->blocksize;
      i2 = p2 - b2*heap->blocksize;
      block2 = heap->blocks[b2];
      if (value < block2->pixels[i2].value) {
         block->pixels[i].pointer = block2->pixels[i2].pointer;
         block->pixels[i].value = block2->pixels[i2].value;
         block = block2;
         b = b2;
         i = i2;
         p = p2;
      }
      else {
         /* we're done */
         break;
      }
   }
   block->pixels[i].pointer = pointer;
   block->pixels[i].value = value;
}

/* This one assumes item p is where the data is at, and moves this item
 * down (away from the root) to sort it */
static void PQT__MoveDown
(
   PQT_priorityqueue heap,
   int p,                    /* item to be sorted */
   PQT_block block,          /* it's in this block */
   int b,                    /* this is the block number */
   int i,                    /* this is the index into the block */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2, b2, i2;       /* for left child */
   PQT_block block2;
   int b3, i3;           /* for right child */
   PQT_block block3;

   p2 = 2*p+1; /* left child. right = 2*p+2 = p2+1 */
   while (p2 < heap->nelements) {
      b2 = p2 / heap->blocksize;
      i2 = p2 - b2*heap->blocksize;
      block2 = heap->blocks[b2];
      if (p2+1 < heap->nelements) {
         if (i2+1 < heap->blocksize) {
            /* same block */
            b3 = b2;
            i3 = i2+1;
            block3 = block2;
         }
         else {
            /* next block */
            b3 = b2+1;
            i3 = 0;
            block3 = heap->blocks[b3];
         }
         if (block3->pixels[i3].value < block2->pixels[i2].value) {
            block2 = block3;
            b2 = b3;
            i2 = i3;
            p2 = p2+1;
         }
      }
      /* compare p to p2 */
      if (block2->pixels[i2].value < value) {
         block->pixels[i].pointer = block2->pixels[i2].pointer;
         block->pixels[i].value = block2->pixels[i2].value;
         block = block2;
         b = b2;
         i = i2;
         p = p2;
         p2 = 2*p+1;
      }
      else {
         /* we're done */
         break;
      }
   }
   block->pixels[i].pointer = pointer;
   block->pixels[i].value = value;
}


PQT_error PQT_Push
(
   PQT_priorityqueue heap,
   void* pointer,
   float value
)
{
   PQT_block block;
   int p, b, i;
   PQT_error retval;

   if (heap->nelements >= heap->blocksize * heap->nblocks) {
      retval = PQT__Enlarge (heap);
      if (retval != PQT_OK) {
         return retval;
      }
   }
   p = heap->nelements;       /* position of insertion */
   b = p / heap->blocksize;   /* block where that position is in */
   i = p - b*heap->blocksize; /* index into block where that position is */
   block = heap->blocks[b];
   heap->nelements++;
   PQT__MoveUp (heap, p, block, b, i, pointer, value);

   return PQT_OK;
}


PQT_error PQT_Pop
(
   PQT_priorityqueue heap,
   void* *pointer,
   float *value
)
{
   PQT_block block = heap->blocks[0];
   int p, b, i;

   if (heap->nelements == 0) {
      return PQT_PROBLEM;
   }
   if (pointer) {
      *pointer = block->pixels[0].pointer;
   }
   if (value) {
      *value = block->pixels[0].value;
   }
   heap->nelements--;
   p = heap->nelements;       /* position of deletion */
   b = p / heap->blocksize;   /* block where that position is in */
   i = p - b*heap->blocksize; /* index into block where that position is */
   block = heap->blocks[b];
   PQT__MoveDown (heap, 0, heap->blocks[0], 0, 0,
                  block->pixels[i].pointer, block->pixels[i].value);

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue heap
)
{
   if (heap->nelements == 0) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
