/*
 * Hierarchical heap: an array of pointers to heaps. The array
 * works like a histogram.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "pqt.h"

#define PQT_HEAPSIZE 128       /* initial number of elements in each heap */
#define PQT_HISTSIZE 1024      /* number of heaps allocated in array */

#define PQT_HIST_MINVAL 0;     /* histogram range */
#define PQT_HIST_MAXVAL 100;   /* histogram range */

typedef struct {
   PQT_pixel *pixels;       /* array with info for each pixel */
   int arraysize;           /* number of elements allocated */
   int nelements;           /* number of elements on the heap */
} PQT__heap, *PQT_heap;

struct PQT__tmp__priorityqueue {
   PQT_heap *heap;          /* array with pointers to heaps */
   int arraysize;           /* number of elements allocated in array */
   float binsize;           /* size of each bin */
   float minval;            /* any value in queue >= minval */
   float maxval;            /* any value in queue < maxval */
   int nelements;           /* number of elements on the queue */
   int minelement;          /* first heap to contain any data */
};


PQT_error PQT__Heap_New
(
   PQT_heap *heap,
   int blocksize
)
{
   PQT_heap tmpheap = 0;

   if (blocksize <= 0) {
      blocksize = PQT_HEAPSIZE;
   }

   /* Allocate & initialise structure */
   tmpheap = (PQT_heap)malloc (sizeof (PQT__heap));
   if (!tmpheap) goto PQT_ERROR;
   tmpheap->pixels = 0;
   tmpheap->nelements = 0;
   tmpheap->arraysize = 0;

   /* Allocate & initialise heap */
   tmpheap->pixels = (PQT_pixel*)malloc (blocksize * sizeof (PQT_pixel));
   if (!tmpheap->pixels) goto PQT_ERROR;
   tmpheap->arraysize = blocksize;

   /* Do the switch */
   *heap = tmpheap;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpheap) {
      free (tmpheap);
   }
   return PQT_PROBLEM;
}


void PQT__Heap_Free
(
   PQT_heap heap
)
{
   if (heap) {
      free (heap->pixels);
      free (heap);
   }
}


static PQT_error PQT__Heap_Enlarge
(
   PQT_heap heap
)
{
   PQT_pixel* tmppixels = 0;

   /* Allocate & initialise heap block */
   tmppixels = (PQT_pixel*)realloc (heap->pixels, 2 * heap->arraysize * sizeof (PQT_pixel));
   if (!tmppixels) return PQT_PROBLEM;
   heap->pixels = tmppixels;
   heap->arraysize *= 2;
   tmppixels = 0;

   return PQT_OK;
}


/* This one assumes item p is where the data is at, and moves this item
 * up (towards the root) to sort it */
static void PQT__Heap_MoveUp
(
   PQT_heap heap,
   int p,                    /* item to be sorted */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2;       /* for parent */

   while (p > 0) {
      p2 = (p-1)/2; /* parent */
      if (value < heap->pixels[p2].value) {
         heap->pixels[p].pointer = heap->pixels[p2].pointer;
         heap->pixels[p].value   = heap->pixels[p2].value;
         p = p2;
      }
      else {
         /* we're done */
         break;
      }
   }
   heap->pixels[p].pointer = pointer;
   heap->pixels[p].value   = value;
}

/* This one assumes item p is where the data is at, and moves this item
 * down (away from the root) to sort it */
static void PQT__Heap_MoveDown
(
   PQT_heap heap,
   int p,                    /* item to be sorted */
   void* pointer,            /* data to be inserted */
   float value               /* data to be inserted */
)
{
   int p2 = 2*p+1;   /* left child. right = 2*p+2 = p2+1 */
   while (p2 < heap->nelements) {
      if (p2+1 < heap->nelements) {
         if (heap->pixels[p2+1].value < heap->pixels[p2].value) {
            p2 = p2+1;
         }
      }
      /* compare p to p2 */
      if (heap->pixels[p2].value < value) {
         heap->pixels[p].pointer = heap->pixels[p2].pointer;
         heap->pixels[p].value   = heap->pixels[p2].value;
         p = p2;
         p2 = 2*p+1;
      }
      else {
         /* we're done */
         break;
      }
   }
   heap->pixels[p].pointer = pointer;
   heap->pixels[p].value = value;
}


PQT_error PQT__Heap_Push
(
   PQT_heap heap,
   void* pointer,
   float value
)
{
   int p;
   PQT_error retval;

   if (heap->nelements >= heap->arraysize) {
      retval = PQT__Heap_Enlarge (heap);
      if (retval != PQT_OK) {
         return retval;
      }
   }
   p = heap->nelements;       /* position of insertion */
   heap->nelements++;
   PQT__Heap_MoveUp (heap, p, pointer, value);

   return PQT_OK;
}


PQT_error PQT__Heap_Pop
(
   PQT_heap heap,
   void* *pointer,
   float *value
)
{
   int p;

   if (heap->nelements == 0) {
      return PQT_PROBLEM;
   }
   if (pointer) {
      *pointer = heap->pixels[0].pointer;
   }
   if (value) {
      *value = heap->pixels[0].value;
   }
   heap->nelements--;
   p = heap->nelements;       /* position of deletion */
   PQT__Heap_MoveDown (heap, 0, heap->pixels[p].pointer, heap->pixels[p].value);

   return PQT_OK;
}


PQT_error PQT_New
(
   PQT_priorityqueue *queue,
   int ignore
)
{
   PQT_priorityqueue tmpqueue = 0;
   int ii;

   /* Allocate & initialise structure */
   tmpqueue = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmpqueue) goto PQT_ERROR;
   tmpqueue->heap = 0;
   tmpqueue->nelements = 0;

   /* Allocate & initialise histogram */
   tmpqueue->heap = (PQT_heap*)malloc (PQT_HISTSIZE * sizeof (PQT__heap));
   if (!tmpqueue->heap) goto PQT_ERROR;
   for (ii=0; ii<PQT_HISTSIZE; ii++) {
      tmpqueue->heap[ii] = 0;
   }
   tmpqueue->arraysize = PQT_HISTSIZE;
   tmpqueue->minval = PQT_HIST_MINVAL;  /* hard-coded!!! */
   tmpqueue->maxval = PQT_HIST_MAXVAL;  /* hard-coded!!! */
   tmpqueue->binsize = (tmpqueue->maxval - tmpqueue->minval) / PQT_HISTSIZE;
   tmpqueue->minelement = PQT_HISTSIZE-1;

   /* Do the switch */
   *queue = tmpqueue;
   tmpqueue = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpqueue) {
      if (tmpqueue->heap) {
         free (tmpqueue->heap);
      }
      free (tmpqueue);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue queue
)
{
   int ii;
   if (queue) {
      for (ii=0; ii<queue->arraysize; ii++) {
         if (queue->heap[ii]) {
            PQT__Heap_Free (queue->heap[ii]);
         }
      }
      free (queue->heap);
      free (queue);
   }
}


PQT_error PQT_Push
(
   PQT_priorityqueue queue,
   void* pointer,
   float value
)
{
   int bin;
   PQT_error retval;

   bin = (int)((value - queue->minval) / queue->binsize);
   if (bin < 0) {
      bin = 0;
   }
   else if (bin >= queue->arraysize) {
      bin = queue->arraysize-1;
   }

   if (! queue->heap[bin]) {
      retval = PQT__Heap_New (&(queue->heap[bin]), 0);
      if (retval) return retval;
   }
   retval = PQT__Heap_Push (queue->heap[bin], pointer, value);
   if (retval) return retval;

   if (queue->minelement > bin) {
      queue->minelement = bin;
   }
   queue->nelements++;

   return PQT_OK;
}


PQT_error PQT__FindMinimumHeap
(
   PQT_priorityqueue queue
)
{
   int bin = queue->minelement + 1; /* if we're here we know minelement is bogus */
   while (bin < queue->arraysize) {
      if (queue->heap[bin] && queue->heap[bin]->nelements > 0) {
         queue->minelement = bin;
         return PQT_OK;
      }
      bin++;
   }
   return PQT_PROBLEM;
}


PQT_error PQT_Pop
(
   PQT_priorityqueue queue,
   void* *pointer,
   float *value
)
{
   int bin;
   PQT_error retval;

   if (queue->nelements == 0) {
      return PQT_PROBLEM;
   }

   bin = queue->minelement;
   if (! queue->heap[bin] || queue->heap[bin]->nelements == 0) {
      retval = PQT__FindMinimumHeap (queue);
      if (retval) return retval;
      bin = queue->minelement;
   }

   retval = PQT__Heap_Pop (queue->heap[bin], pointer, value);
   if (retval) return retval;

   queue->nelements--;

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue queue
)
{
   if (queue->nelements == 0) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
