/*
 * Ladder queue: Multi-scale queue with a sorted linked list at the
 * bottom, multiple rungs with unsorted buckets at intermediate levels,
 * and an unsorted list at the top.
 *
 * Implemented according to Tang et al., ACM Trans. on Modeling and
 * Computer Simulation 15(3):175-204, 2005.
 * One minor difference: In the paper it is suggested to increase the
 * length of bottom in the case there are more elements with the same
 * value than fit in bottom. Here we move only as many as fit in bottom,
 * leaving the remainder in the bucket. This bucket then also receives
 * potential new items with that same value.
 *
 * (c)2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pqt.h"

#define PQT_LQ_DEBUG 0

#if PQT_LQ_DEBUG
   #define PQT_LQ_ASSERT(v,s) if (~(v)) { printf (s); fflush (stdout); return PQT_PROBLEM; }
   #define PQT_LQ_PRINT(s)    { printf (s); fflush (stdout); }
   #define PQT_LQ_PRINT2(s)   { printf (s); fflush (stdout); }
   #define PQT_LQ_PRINT3(s)   { /*printf (s);*/ }
#else
   #define PQT_LQ_ASSERT(v,s) {}
   #define PQT_LQ_PRINT(s)    {}
   #define PQT_LQ_PRINT2(s)   {}
   #define PQT_LQ_PRINT3(s)   {}
#endif

#define PQT_BOTSIZE 50  /* maximum number of elements in the bottom list */
#define PQT_LISTSIZE 50 /* initial size of each list element */
#define PQT_MAXRUNGS 8  /* maximum number of rungs in the ladder */
#define PQT_MAX_INIT_TOP 900000 /* maximum number of elements in top during the first push cycle */

typedef struct PQT__tmp__node {
   PQT_pixel pixel;              /* pixel data */
   struct PQT__tmp__node *next;  /* next node in linked list */
} PQT__node, *PQT_node;

typedef struct PQT__tmp__list {
   PQT_pixel* pixels;            /* array of pixel data */
   int nelements;                /* number of pixels in list */
   int listsize;                 /* length of array allocated for list */
} PQT__list, *PQT_list;

typedef struct PQT__tmp__rung {
   PQT_list* buckets;            /* array of buckets, each one has a list */
   int nelements;                /* number of pixels in list */
   int nbuckets;                 /* number of buckets in rung */
   double bucketwidth;           /* width of buckets in run */
   double start;                 /* lowest priority value of first bucket */
   double current;               /* threshold of priority value to be stored in rung */
   int curbuck;                  /* next bucket to be emptied */
} PQT__rung, *PQT_rung;

struct PQT__tmp__priorityqueue {
   PQT_list top;                 /* top, unsorted list */
   double topmaxval;             /* maximum value in top */
   double topminval;             /* minimum value in top */
   double topstart;              /* threshold of priority values to be stored in top */
   
   PQT__rung rungs[PQT_MAXRUNGS];/* array of pointers to rungs */
   int nrungs;                   /* number of rungs */

   PQT__node bottomarray[PQT_BOTSIZE]; /* array with nodes for bottom, a sorted linked list */
   PQT_node bottom;              /* pointer to first element */
   PQT_node freelist;            /* pointer to first free block */
   int nbotelem;                 /* number of elements in bottom */

   int nelements;                /* number of elements on the queue */
};


PQT_error PQT__list_New
(
   PQT_list *list
)
{
   PQT_list tmplist = 0;

   /* Allocate & initialise structure */
   tmplist = (PQT_list)malloc (sizeof (PQT__list));
   if (!tmplist) goto PQT_ERROR;
   tmplist->pixels = 0;
   tmplist->nelements = 0;
   tmplist->listsize = 0;

   /* Allocate & initialise list */
   tmplist->pixels = (PQT_pixel*)malloc (PQT_LISTSIZE * sizeof (PQT_pixel));
   if (!tmplist->pixels) goto PQT_ERROR;
   tmplist->listsize = PQT_LISTSIZE;

   /* Do the switch */
   *list = tmplist;
   tmplist = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmplist) {
      free (tmplist);
   }
   return PQT_PROBLEM;
}


void PQT__list_Free
(
   PQT_list list
)
{
   if (list) {
      free (list->pixels);
      free (list);
   }
}


static PQT_error PQT__list_Enlarge
(
   PQT_list list
)
{
   PQT_pixel* tmppixels = 0;

   /* Allocate & initialise list block */
   tmppixels = (PQT_pixel*)realloc (list->pixels, 2 * list->listsize * sizeof (PQT_pixel));
   if (!tmppixels) return PQT_PROBLEM;
   list->pixels = tmppixels;
   list->listsize *= 2;
   tmppixels = 0;

   return PQT_OK;
}


static PQT_error PQT__list_Append
(
   PQT_list list,
   void* pointer,
   float value
)
{
   PQT_error retval;

   if (list->nelements>=list->listsize) {
      retval = PQT__list_Enlarge (list);
      if (retval) return retval;
   }
   list->pixels[list->nelements].pointer = pointer;
   list->pixels[list->nelements].value = value;
   list->nelements++;
   
   return PQT_OK;
}


static PQT_error PQT__rung_New
(
   PQT_rung rung,
   int nbuckets
)
{
   int ii;
   PQT_error retval;

   PQT_LQ_PRINT("### Allocating rung.\n")

   /* structure is allocated, just need to allocate buckets. */
   rung->buckets = (PQT_list*)calloc (nbuckets, sizeof (PQT_list));
   if (!rung->buckets) return PQT_PROBLEM;
   rung->nbuckets = nbuckets;
   rung->nelements = 0;
   rung->curbuck = 0;

   for (ii=0; ii<nbuckets; ii++) {
      retval = PQT__list_New (&(rung->buckets[ii]));
      if (retval) goto PQT_ERROR;
   }   

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (rung->buckets) {
      for (ii=0; ii<nbuckets; ii++) {
         if (rung->buckets[ii]) {
            PQT__list_Free (rung->buckets[ii]);
         }
      }   
      free (rung->buckets);
   }
   rung->nbuckets = 0;
   rung->buckets = 0;
   return PQT_PROBLEM;
}


static void PQT__rung_Free
(
   PQT_rung rung
)
{
   int ii;

   PQT_LQ_PRINT("### Freeing rung.\n")

   /* just need to free buckets. */
   for (ii=0; ii<rung->nbuckets; ii++) {
      if (rung->buckets[ii]) {
         PQT__list_Free (rung->buckets[ii]);
      }
   }   

   free (rung->buckets);
   rung->nbuckets = 0;
   rung->buckets = 0;
}


static PQT_error PQT__rung_Enlarge
(
   PQT_rung rung,
   int nbuckets
)
{
   /* This function will only ever be called for the first rung! */
   PQT_list* tmpbuckets = 0;
   int ii;
   PQT_error retval;

   PQT_LQ_PRINT("### Enlarging rung.\n")

   /* structure is allocated, just need to allocate buckets. */
   if (nbuckets < 2*rung->nbuckets) {
      nbuckets = 2*rung->nbuckets;
   }
   tmpbuckets = (PQT_list*)realloc (rung->buckets, nbuckets * sizeof (PQT_list));
   if (!tmpbuckets) return PQT_PROBLEM;
   for (ii=rung->nbuckets; ii<nbuckets; ii++) {
      tmpbuckets[ii] = 0;
   }   
   for (ii=rung->nbuckets; ii<nbuckets; ii++) {
      retval = PQT__list_New (&(tmpbuckets[ii]));
      if (retval) goto PQT_ERROR;
   }   

   /* switch */
   rung->buckets = tmpbuckets;
   rung->nbuckets = nbuckets;
   tmpbuckets = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpbuckets) {
      for (ii=0; ii<nbuckets; ii++) {
         if (tmpbuckets[ii]) {
            PQT__list_Free (tmpbuckets[ii]);
         }
      }   
      free (tmpbuckets);
   }
   return PQT_PROBLEM;
}


PQT_error PQT_New
(
   PQT_priorityqueue *queue,
   int ignore
)
{
   PQT_priorityqueue tmpqueue = 0;
   int ii;
   PQT_error retval;

   /* Allocate & initialise structure */
   tmpqueue = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmpqueue) goto PQT_ERROR;
   tmpqueue->top = 0;
   tmpqueue->nrungs = 0;
   tmpqueue->nelements = 0;
   for (ii=0; ii<PQT_MAXRUNGS; ii++) {
      tmpqueue->rungs[ii].buckets = 0;
      tmpqueue->rungs[ii].nbuckets = 0;
      tmpqueue->rungs[ii].nelements = 0;
   }
   
   /* Initialise bottom */
   tmpqueue->bottom = 0;
   tmpqueue->nbotelem = 0;
   tmpqueue->freelist = tmpqueue->bottomarray;
   for (ii=0; ii<(PQT_BOTSIZE-1); ii++) {
      tmpqueue->bottomarray[ii].next = &(tmpqueue->bottomarray[ii+1]);
   }
   tmpqueue->bottomarray[PQT_BOTSIZE-1].next = 0;

   /* Allocate & initialise top */
   retval = PQT__list_New (&(tmpqueue->top));
   if (retval) goto PQT_ERROR;
   tmpqueue->topmaxval = 0;   /* these values are initialized and updated as we push data */
   tmpqueue->topminval = 0;   /*    onto the queue, and remain "floating" as long as */
   tmpqueue->topstart = 0;    /*    nbotelem==0 and nrungs==0. */

   /* Allocate 5 rungs */
   /* (in the paper they start with allocating 5) */
   for (ii=0; ii<5; ii++) {
      retval = PQT__rung_New (&(tmpqueue->rungs[ii]), PQT_BOTSIZE);
      if (retval) goto PQT_ERROR;
   }

   /* Do the switch */
   *queue = tmpqueue;
   tmpqueue = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmpqueue) {
      if (tmpqueue->top) {
         PQT__list_Free (tmpqueue->top);
      }
      for (ii=0; ii<5; ii++) {
         if (tmpqueue->rungs[ii].buckets) {
            PQT__rung_Free (&(tmpqueue->rungs[ii]));
         }
      }
      free (tmpqueue);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue queue
)
{
   int ii;
   if (queue) {
      for (ii=0; ii<PQT_MAXRUNGS; ii++) {
         if (queue->rungs[ii].buckets) {
            PQT__rung_Free (&(queue->rungs[ii]));
         }
      }
      if (queue->top) {
         PQT__list_Free (queue->top);
      }
      free (queue);
   }
}


static void PQT__Sort_into_Bottom
(
   PQT_priorityqueue queue,
   PQT_pixel px
)
{
   PQT_node new = queue->freelist;
   queue->freelist = queue->freelist->next;
   new->pixel = px;
   if (queue->bottom==0) {
      new->next = 0;
      queue->bottom = new;
   }
   else if (queue->bottom->pixel.value>px.value) {
      new->next = queue->bottom;
      queue->bottom = new;
   }
   else {
      PQT_node cur = queue->bottom;
      while (cur->next!=0 && cur->next->pixel.value<=px.value) {
         cur = cur->next;
      }
      new->next = cur->next;
      cur->next = new;
   }
   queue->nbotelem++;  
}


static PQT_error PQT__New_Rung_from_Bottom
(
   PQT_priorityqueue queue
)
{
   int rung, bin, ii;
   double maxval, minval;
   PQT_node cur;
   PQT_error retval;

   PQT_LQ_PRINT2("--- --- ---Creating a new rung from bottom (#1)\n")

   /* Create new rung */
   if (queue->nrungs >= PQT_MAXRUNGS) return PQT_PROBLEM;
   rung = queue->nrungs;
   if (queue->rungs[rung].nbuckets == 0) {
      retval = PQT__rung_New (&(queue->rungs[rung]), PQT_BOTSIZE);
      if (retval) return retval;
   } /* If it exists, it will have at least PQT_BOTSIZE buckets allocated */
   PQT_LQ_ASSERT( queue->rungs[rung].nbuckets>=PQT_BOTSIZE, "ASSERTION FAILED!!! (#1) Rung has too few buckets.\n" )
   queue->nrungs++;
   minval = queue->bottom->pixel.value;
   if (rung == 0) {
      maxval = queue->topstart;
      queue->rungs[rung].bucketwidth = (maxval - minval) / queue->rungs[rung].nbuckets * 1.001;
   }
   else {
      queue->rungs[rung].bucketwidth = queue->rungs[rung-1].bucketwidth / queue->rungs[rung].nbuckets * 1.001;
   }
   if (queue->rungs[rung].bucketwidth == 0) {
      queue->rungs[rung].bucketwidth = 1e-9; /* Avoid zero-width buckets */
   }
   queue->rungs[rung].current = queue->rungs[rung].start = minval;
   queue->rungs[rung].curbuck = 0;
   
   /* Fill it from bottom linked list */
   cur = queue->bottom;
   while (cur) {
      bin = (cur->pixel.value - minval) / queue->rungs[rung].bucketwidth;
      PQT_LQ_ASSERT( bin < queue->rungs[rung].nbuckets, "ASSERTION FAILED!!! (#2) Pushing in a bucket past the rung length.\n" )
      PQT__list_Append (queue->rungs[rung].buckets[bin], cur->pixel.pointer, cur->pixel.value);
      cur = cur->next;
   }
   queue->rungs[rung].nelements = queue->nbotelem;

   /* Reset bottom */
   queue->bottom = 0;
   queue->nbotelem = 0;
   queue->freelist = queue->bottomarray;
   for (ii=0; ii<(PQT_BOTSIZE-1); ii++) {
      queue->bottomarray[ii].next = &(queue->bottomarray[ii+1]);
   }
   queue->bottomarray[PQT_BOTSIZE-1].next = 0;

   return PQT_OK;   
}


static PQT_error PQT__New_Rung_from_List
(
   PQT_priorityqueue queue,
   PQT_list list,
   double minval,
   double maxval
)
{
   int rung, bin, ii, nbuckets;
   double width;
   PQT_error retval;

   PQT_LQ_PRINT2("--- --- ---Creating a new rung from a list (#2)\n")

   /* Create new rung */
   if (queue->nrungs >= PQT_MAXRUNGS) return PQT_PROBLEM;
   rung = queue->nrungs;
   if (rung == 0) {
      nbuckets = list->nelements;
   }
   else {
      nbuckets = PQT_BOTSIZE;
   }
   width = (maxval - minval) / nbuckets * 1.001;
   if (width<=0) {
      width = 1e-9; /* Avoid zero-width buckets */
   }
   if (nbuckets < PQT_BOTSIZE) {
      nbuckets = PQT_BOTSIZE;
   }
   if (queue->rungs[rung].nbuckets == 0) {
      retval = PQT__rung_New (&(queue->rungs[rung]), nbuckets);
      if (retval) return retval;
   }
   else if (queue->rungs[rung].nbuckets < nbuckets) {
      retval = PQT__rung_Enlarge (&(queue->rungs[rung]), nbuckets);
      if (retval) return retval;
   }
   queue->rungs[rung].bucketwidth = width;
   queue->rungs[rung].current = queue->rungs[rung].start = minval;
   queue->rungs[rung].curbuck = 0;
   queue->nrungs++;
   
   /* Fill it from list */
   for (ii=0; ii<list->nelements; ii++) {
      bin = (list->pixels[ii].value - minval) / queue->rungs[rung].bucketwidth;
      PQT__list_Append (queue->rungs[rung].buckets[bin], list->pixels[ii].pointer, list->pixels[ii].value);
   }
   queue->rungs[rung].nelements = list->nelements;

   return PQT_OK;   
}


PQT_error PQT_Push
(
   PQT_priorityqueue queue,
   void* pointer,
   float value
)
{
   int rung, bin;
   PQT_error retval;

   if ((queue->nbotelem==0 && queue->nrungs==0) || (value >= queue->topstart)) {
      /* Push into top */
      PQT_LQ_PRINT3("---Pushing into top (#3)\n")
      retval = PQT__list_Append (queue->top, pointer, value);
      if (retval) return retval;
      if (queue->topmaxval < value)
         queue->topmaxval = value;
      if (queue->topminval > value)
         queue->topminval = value;
      if (queue->topstart > value)
         /* only happens when nbotelem==0 && nrungs==0 */
         queue->topstart = value;
      if (queue->nbotelem==0 && queue->nrungs==0 && queue->top->nelements > PQT_MAX_INIT_TOP) {
         /* This is to avoid the top list to become too large when initialising
          * the queue in the tests we do here. This part should not really be
          * here for normal usage. It is not common to put a million elements
          * into a queue without taking any out along the way. As soon as we've
          * got 1 million elements, we transfer the stuff into the ladder just
          * as if we'd be popping a value. Following values will end up in top or
          * in the ladder, depending on their value.
          */
         PQT_LQ_PRINT2("--- ---Transferring top into ladder (#4)\n")
         retval = PQT__New_Rung_from_List (queue, queue->top, queue->topminval, queue->topmaxval);
         if (retval) return retval;
         queue->top->nelements = 0;
         queue->topstart = queue->rungs[0].start + queue->rungs[0].nbuckets * queue->rungs[0].bucketwidth;
         queue->topminval = queue->topmaxval = queue->topstart;
      }
   }
   else {
      /* Find the run to push in */
      PQT_boolean done = PQT_FALSE;
      rung = 0;
      while ((rung < queue->nrungs) && (queue->rungs[rung].current > value)) {
         rung++;
      }
      if (rung >= queue->nrungs) {
         /* Push into bottom */
         if (queue->nbotelem < PQT_BOTSIZE) {
            PQT_pixel new;
            PQT_LQ_PRINT3("---Pushing into bottom (#5)\n")
            new.pointer = pointer;
            new.value = value;
            PQT__Sort_into_Bottom (queue, new);
            done = PQT_TRUE;
         }
         else {
            /* Make a new rung, copy all of bottom into it; the item will be pushed
               later into this new rung */
            retval = PQT__New_Rung_from_Bottom (queue);
            if (retval) return retval;
            rung = queue->nrungs-1;
         }
      }
      if (!done) {
         /* Push in this rung */
         PQT_LQ_PRINT3("---Pushing into ladder (#6)\n")
         bin = ( value - queue->rungs[rung].start ) / queue->rungs[rung].bucketwidth;
         PQT_LQ_ASSERT( bin < queue->rungs[rung].nbuckets, "ASSERTION FAILED!!! (#3) Pushing in a bucket past the rung length.\n" )
         PQT_LQ_ASSERT( bin >= queue->rungs[rung].curbuck, "ASSERTION FAILED!!! (#4) Pushing in a bucket that is already empty.\n" )
            /* This last ASSERT showed the problem with round-off errors that made
               me use double everywhere (except the input data is still float).
               When "value" is too close to "current", bin could turn out to be
               just slightly below the correct integer value, and be rounded down
               to the wrong bin. */
         retval = PQT__list_Append (queue->rungs[rung].buckets[bin], pointer, value);
         if (retval) return retval;
         queue->rungs[rung].nelements++;
      }
   }

   queue->nelements++;

   return PQT_OK;
}


PQT_error PQT_Pop
(
   PQT_priorityqueue queue,
   void* *pointer,
   float *value
)
{
   PQT_node freenode;
   int rung, bin, ii;
   PQT_boolean done, empty;
   PQT_error retval;
   
   if (queue->nbotelem == 0) {
      if (queue->nrungs == 0) {
         /* Transfer all of the top data into the ladder */
         PQT_LQ_PRINT2("--- ---Transferring top into ladder (#7)\n")
         retval = PQT__New_Rung_from_List (queue, queue->top, queue->topminval, queue->topmaxval);
         if (retval) return retval;
         queue->top->nelements = 0;
         queue->topstart = queue->rungs[0].start + queue->rungs[0].nbuckets * queue->rungs[0].bucketwidth;
         queue->topminval = queue->topmaxval = queue->topstart;
      }
      /* Fill bottom sorted list from one bucket in the ladder */
      done = PQT_FALSE;
      while (!done) {
         rung = queue->nrungs-1;
         PQT_LQ_ASSERT( queue->rungs[rung].nelements != 0, "ASSERTION FAILED!!! (#5) Found an empty rung.\n" )
         for (bin=queue->rungs[rung].curbuck; bin<queue->rungs[rung].nbuckets; bin++) {
            if (queue->rungs[rung].buckets[bin]->nelements > 0) {
               /* This is our bucket! */
               if (queue->rungs[rung].buckets[bin]->nelements > PQT_BOTSIZE) {
                  /* Spawn a new rung */
                  double minval, maxval;
                  PQT_list list = queue->rungs[rung].buckets[bin];
                  minval = queue->rungs[rung].start + bin * queue->rungs[rung].bucketwidth;
                  maxval = minval + queue->rungs[rung].bucketwidth;
                  empty = PQT_FALSE;
                  if (queue->rungs[rung].buckets[bin]->nelements == queue->rungs[rung].nelements) {
                     /* This is the last bucket with data in the rung. */
                     /* There's two options: either this just happens to be the last
                        bucket, or we've got a big collection of equally-valued elements.
                        In the latter case, we've got to do something special to avoid
                        infinite recursion. */
                     PQT_boolean allequal = PQT_TRUE;
                     float value = list->pixels[0].value;
                     for (ii=1; ii<list->nelements; ii++) {
                        if (list->pixels[ii].value != value) {
                           allequal = PQT_FALSE;
                           break;
                        }
                     }
                     if (allequal) {
                        /* Sort part of this data into bottom, keep the rest in the bucket. */
                        PQT_LQ_PRINT2("--- ---Transferring bucket partially into bottom (#13)\n")
                        for (ii=0; ii<PQT_BOTSIZE; ii++) {
                           PQT__Sort_into_Bottom (queue, list->pixels[ii]);
                           /* could be done more efficiently, we know they all have the same value... */
                        }
                        list->nelements -= PQT_BOTSIZE;
                        queue->rungs[rung].nelements -= PQT_BOTSIZE;
                        memmove (list->pixels, (list->pixels+PQT_BOTSIZE), list->nelements*sizeof(PQT_pixel));
                        done = PQT_TRUE;
                     }
                     else {
                        /* Delete this rung first, saving our list */
                        retval = PQT__list_New (&(queue->rungs[rung].buckets[bin]));
                        if (retval) return retval;
                        queue->rungs[rung].nelements = 0;
                        queue->nrungs--;
                        empty = PQT_TRUE;
                     }
                  }
                  if (!done) {
                     PQT_LQ_PRINT2("--- ---Transferring bucket into new rung (#8)\n")
                     retval = PQT__New_Rung_from_List (queue, list, minval, maxval);
                     if (retval) return retval;
                     if (empty) {
                        /* Delete the list, it's no longer linked anywhere */
                        PQT__list_Free (list);
                        PQT_LQ_PRINT2("--- --- ---Deleted a rung (#9)\n")
                     }
                     else {
                        /* Update counters */
                        queue->rungs[rung].nelements -= list->nelements;
                        list->nelements = 0;
                        queue->rungs[rung].curbuck = bin+1;
                        queue->rungs[rung].current = queue->rungs[rung].start + (bin+1) * queue->rungs[rung].bucketwidth;
                     }
                     /* We don't set "done", so that a new iteration of the while loop
                        takes data from the new rung into bottom. */
                  }
               }
               else {
                  /* Sort this data into bottom */
                  PQT_LQ_PRINT2("--- ---Transferring bucket into bottom (#10)\n")
                  for (ii=0; ii<queue->rungs[rung].buckets[bin]->nelements; ii++) {
                     PQT__Sort_into_Bottom (queue, queue->rungs[rung].buckets[bin]->pixels[ii]);
                  }
                  queue->rungs[rung].nelements -= queue->rungs[rung].buckets[bin]->nelements;
                  queue->rungs[rung].buckets[bin]->nelements = 0;
                  if (queue->rungs[rung].nelements == 0) {
                     queue->nrungs--;
                     PQT_LQ_PRINT2("--- --- ---Deleted a rung (#11)\n")
                  }
                  else {
                     queue->rungs[rung].curbuck = bin+1;
                     queue->rungs[rung].current = queue->rungs[rung].start + (bin+1) * queue->rungs[rung].bucketwidth;
                  }
                  done = PQT_TRUE;
               }
               break;
            }
         }
         PQT_LQ_ASSERT( bin < queue->rungs[rung].nbuckets, "ASSERTION FAILED!!! (#6) Found a non-empty rung where all buckets are empty.\n" )
      }
   }

   /* Take data from first element in bottom */
   PQT_LQ_PRINT3("---Popping from bottom (#12)\n")
   if (pointer) {
      *pointer = queue->bottom->pixel.pointer;
   }
   if (value) {
      *value = queue->bottom->pixel.value;
   }
   /* Unlink first element and add to free list */
   freenode = queue->bottom;
   queue->bottom = queue->bottom->next;
   freenode->next =  queue->freelist;
   queue->freelist = freenode;

   queue->nbotelem--;
   queue->nelements--;

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue queue
)
{
   if (queue->nelements == 0) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue queue
)
{
   return queue->nelements;
}



void PQT_Print
(
   PQT_priorityqueue queue
)
{
   int ii;

   printf( "Queue status:\n   No. Elements: %d\n", queue->nelements );
   printf( "   Top:\n      No. Elements: %d (%d allocated)\n", queue->top->nelements, queue->top->listsize );
   printf( "      Min=%f, max=%f, start=%f\n", queue->topminval, queue->topmaxval, queue->topstart );
   printf( "   Ladder:\n      No. Rungs: %d\n", queue->nrungs );
   for (ii=0; ii<queue->nrungs; ii++) {
      printf( "    Rung #%d:\n      No. Elements: %d\n", ii, queue->rungs[ii].nelements );
      printf( "      No. Buckets: %d, current bucket: %d\n", queue->rungs[ii].nbuckets, queue->rungs[ii].curbuck );
      printf( "      Width=%f, start=%f, current=%f\n", queue->rungs[ii].bucketwidth, queue->rungs[ii].start, queue->rungs[ii].current );
   }
   printf( "   Bottom:\n      No. Elements: %d\n", queue->nbotelem );
}
