/*
 * A left-leaning red-black tree, the nodes are allocated in large arrays.
 * Using bits of code by Robert Sedgewick (2008).
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "pqt.h"

#define PQT_BLOCKSIZE 1024
#define PQT_NUMBEROFBLOCKS 128

typedef struct PQT__tmp__node {
   PQT_pixel pixel;               /* pixel data */
   struct PQT__tmp__node *left;   /* left child node */
   struct PQT__tmp__node *right;  /* right child node */
   PQT_boolean red;               /* true for red node, false for black node */
} PQT__node, *PQT_node;

struct PQT__tmp__priorityqueue {
   PQT_node root;           /* pointer to the root */
   PQT_node freelist;       /* pointer to the list of free nodes; node->left points to the next empty node in the list */
   int nelements;           /* number of elements on the queue */
   int blocksize;           /* number of nodes allocated in each block */
   int nblocks;             /* number of blocks allocated */
   int arraysize;           /* number of elements in pointer array */
   PQT_node *blocks;        /* array of pointers to the blocks */
};


static PQT_error PQT__NewBlock
(
   PQT_priorityqueue tree
)
{
   PQT_node block;
   int ii;

   /* Check & extend pointer array */
   if (tree->nblocks >= tree->arraysize) {
      PQT_node* tmpblocks = 0;
      tmpblocks = (PQT_node*)realloc (tree->blocks, 2 * tree->arraysize * sizeof (PQT_node));
      if (!tmpblocks) {
         return PQT_PROBLEM;
      }
      tree->blocks = tmpblocks;
      tree->arraysize *= 2;
   }

   /* Allocate & initialise tree block */
   block = (PQT_node)malloc (tree->blocksize * sizeof (PQT__node));
   if (!block) {
      return PQT_PROBLEM;
   }
   tree->blocks[tree->nblocks] = block;
   tree->nblocks++;
   for (ii=0; ii<tree->blocksize; ii++) {
      block[ii].left = &(block[ii+1]);
   }
   block[tree->blocksize-1].left = tree->freelist;
   tree->freelist = block;

   return PQT_OK;
}


static PQT_error PQT__NewNode
(
   PQT_node *node,
   PQT_priorityqueue tree,
   void* pointer,
   float value,
   PQT_boolean red
)
{
   PQT_node tmpnode;

   if (!tree->freelist) {
      PQT_error retval = PQT__NewBlock (tree);
      if (retval != PQT_OK) {
         return retval;
      }
   }

   tmpnode = tree->freelist;
   tree->freelist = tree->freelist->left;
   tmpnode->left  = 0;
   tmpnode->right = 0;
   tmpnode->pixel.pointer = pointer;
   tmpnode->pixel.value   = value;
   tmpnode->red = red;
   *node = tmpnode;

   return PQT_OK;
}


static void PQT__FreeNode
(
   PQT_node node,
   PQT_priorityqueue tree
)
{
   node->left = tree->freelist;
   tree->freelist = node;
}


PQT_error PQT_New
(
   PQT_priorityqueue *tree,
   int blocksize
)
{
   PQT_priorityqueue tmptree = 0;

   if (blocksize <= 0) {
      blocksize = PQT_BLOCKSIZE;
   }

   tmptree = (PQT_priorityqueue)malloc (sizeof (PQT__priorityqueue));
   if (!tmptree) goto PQT_ERROR;
   tmptree->nelements = 0;
   tmptree->root = 0;
   tmptree->blocksize = blocksize;
   tmptree->nblocks = 0;
   tmptree->freelist = 0;
   tmptree->blocks = (PQT_node*)malloc (PQT_NUMBEROFBLOCKS * sizeof (PQT_node*));
   if (!tmptree->blocks) goto PQT_ERROR;
   tmptree->arraysize = PQT_NUMBEROFBLOCKS;

   *tree = tmptree;
   tmptree = 0;

   return PQT_OK;

PQT_ERROR:
   /* In case of error, undo all allocations */
   if (tmptree) {
      if (tmptree->blocks) {
         free (tmptree->blocks);
      }
      free (tmptree);
   }
   return PQT_PROBLEM;
}


void PQT_Free
(
   PQT_priorityqueue tree
)
{
   int ii;

   if (tree) {
      if (tree->blocks) {
         for (ii=0; ii<tree->nblocks; ii++) {
            free (tree->blocks[ii]);
         }
         free (tree->blocks);
      }
      free (tree);
   }
}


static void PQT__Insert
(
   PQT_node *node,
   PQT_node newnode
)
{
   if (!(*node)) {
      *node = newnode;
   }
   else {
      /* Equal values inserted to the right -- as a minimum-pop queue this is fifo for equal-valued items */
      if ((*node)->pixel.value <= newnode->pixel.value) {
         /* to the right child */
         PQT__Insert (&((*node)->right), newnode);
      }
      else {
         /* to the left child */
         PQT__Insert (&((*node)->left), newnode);
      }
      if ((*node)->right && (*node)->right->red) {
         /*PQT__RotateLeft (node);*/
         PQT_node y = *node;
         PQT_node x = y->right;
         y->right = x->left;
         x->left = y;
         *node = x;
         x->red = y->red;
         y->red = PQT_TRUE;
      }
      if ((*node)->left && (*node)->left->red && (*node)->left->left && (*node)->left->left->red) {
         /*PQT__RotateRight (node);*/
         PQT_node y = *node;
         PQT_node x = y->left;
         y->left = x->right;
         x->right = y;
         *node = x;
         x->red = y->red;
         y->red = PQT_TRUE;
      }
      if ((*node)->left && (*node)->left->red && (*node)->right && (*node)->right->red) {
         /*PQT__FlipColor (*node);*/
         (*node)->red        = !(*node)->red;
         (*node)->left->red  = !(*node)->left->red;
         (*node)->right->red = !(*node)->right->red;
      }
   }
}


PQT_error PQT_Push
(
   PQT_priorityqueue tree,
   void* pointer,
   float value
)
{
   PQT_node newnode;
   PQT_error retval;

   retval = PQT__NewNode (&newnode, tree, pointer, value, PQT_TRUE); /* insert red node */
   if (retval != PQT_OK) {
      return retval;
   }

   if (tree->root) {
      PQT__Insert (&(tree->root), newnode);
   }
   else {
      tree->root = newnode;
   }
   tree->nelements++;

   return PQT_OK;
}


static void PQT__DeleteMin
(
   PQT_node *node,
   PQT_priorityqueue tree,
   PQT_pixel *pixel
)
{
   if ((*node)->left) {
      if (!((*node)->left->red) && !((*node)->left->left && (*node)->left->left->red)) {
         /*PQT__FlipColor (*node);*/
         (*node)->red        = !(*node)->red;
         (*node)->left->red  = PQT_TRUE;
         (*node)->right->red = !(*node)->right->red;
         if ((*node)->right->left && (*node)->right->left->red) {
            /*PQT__RotateRight (&((*node)->right));*/
            /*PQT__RotateLeft (node);*/
            /*PQT__FlipColor (*node);*/
            PQT_node w = *node;
            PQT_node y = w->right;
            PQT_node x = y->left;
            y->left = x->right;
            x->right = y;
            w->right = x->left;
            x->left = w;
            *node = x;
            x->red = !w->red;
            w->red = PQT_FALSE;
            y->red = PQT_FALSE;
         }
      }
      PQT__DeleteMin (&((*node)->left), tree, pixel);
      if ((*node)->right && (*node)->right->red) {
         /*PQT__RotateLeft (node);*/
         PQT_node y = *node;
         PQT_node x = y->right;
         y->right = x->left;
         x->left = y;
         *node = x;
         x->red = y->red;
         y->red = PQT_TRUE;
      }
      if ((*node)->left && (*node)->left->red && (*node)->left->left && (*node)->left->left->red) {
         /*PQT__RotateRight (node);*/
         PQT_node y = *node;
         PQT_node x = y->left;
         y->left = x->right;
         x->right = y;
         *node = x;
         x->red = y->red;
         y->red = PQT_TRUE;
      }
      if ((*node)->left && (*node)->left->red && (*node)->right && (*node)->right->red) {
         /*PQT__FlipColor (*node);*/
         (*node)->red        = !(*node)->red;
         (*node)->left->red  = PQT_FALSE;
         (*node)->right->red = PQT_FALSE;
      }
   }
   else {
      PQT_node tmp = 0;
      *pixel = (*node)->pixel;         /* copies data, not a pointer */
      tmp = (*node)->right;
      PQT__FreeNode (*node, tree);
      *node = tmp;
   }
}


PQT_error PQT_Pop
(
   PQT_priorityqueue tree,
   void* *pointer,
   float *value
)
{
   PQT_pixel pixel;

   if (!tree->root) {
      return PQT_PROBLEM;
   }

   PQT__DeleteMin (&(tree->root), tree, &pixel);
   tree->nelements--;
   if (pointer) {
      *pointer = pixel.pointer;
   }
   if (value) {
      *value = pixel.value;
   }

   return PQT_OK;
}


PQT_boolean PQT_IsEmpty
(
   PQT_priorityqueue tree
)
{
   if (!tree->root) {
      return PQT_TRUE;
   }
   else {
      return PQT_FALSE;
   }
}


int PQT_NElements
(
   PQT_priorityqueue heap
)
{
   return heap->nelements;
}


void PQT_Print ( PQT_priorityqueue queue ) {}
