/*
 * Testing the various priority queues
 *
 * This program:
 *  - generates a large 3D grey-value image
 *  - generates a seed image with several differently-labeled seeds
 *  - runs a region growing algorithm, in which pixels are added to
 *    the regions based on the cumulative grey-value (this is
 *    related to the grey-weighted distance transform).
 *  - the function is timed to show the effect of using one or the
 *    other priority queue.
 *
 * This program uses DIPlib and contains a stripped-down version of
 * the dip_GrowRegionsWeighted() function in DIPlib.
 *
 * (c)2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "pqt.h"
#include "diplib.h"
#include "dip_regions.h"
#include "dip_neighbourlist.h"
#include "dip_coordsindx.h"
#include "dip_noise.h"
#include "dip_linear.h"
#include "dip_point.h"
#include "dip_math.h"
#include "dip_manipulation.h"

#define DEBUG 0


/* Returns DIP_TRUE if a pixel in the list is background and has the mask set */
static dip_Boolean dip__PixelHasBackgroundNeighbour
(
   dip_sint32 *ptr,
   dip_CoordinateArray NLcoords,
   dip_IntegerArray NLindices,
   dip_IntegerArray coords,
   dip_IntegerArray imsz
)
{
   dip_int ii, jj, pos;
   dip_Boolean skip;
   dip_int ndims = imsz->size;
   dip_int nneighbours = NLcoords->size;

   for (jj=0; jj<nneighbours; jj++) {
      skip = DIP_FALSE;
      for (ii=0; ii<ndims; ii++) {
         pos = coords->array[ii] + NLcoords->array[ii][jj];
         if ((pos < 0) || (pos >= imsz->array[ii])) {
            skip = DIP_TRUE;
            break;
         }
      }
      if (!skip) {
         if (*(ptr + NLindices->array[jj]) == 0) {
            return( DIP_TRUE );
         }
      }
   }
   return( DIP_FALSE );
}

/*
 * The thing (In DIPlib this is compiled multiple times for various input data types
 */
static dip_Error dip__GrowRegionsWeighted
(
   dip_sint32 *labels,
   dip_sfloat *grey,
   dip_sfloat *distance,
   dip_IntegerArray imsz,
   PQT_priorityqueue pqueue,
   dip_IntegerArray labstride,
   dip_IntegerArray greystride,
   dip_IntegerArray diststride,
   dip_IntegerArray NLindices_lab,
   dip_IntegerArray NLindices_grey,
   dip_IntegerArray NLindices_dist,
   dip_CoordinateArray NLcoords,
   dip_FloatArray NLdistances
)
{
   DIP_FNR_DECLARE("dip__GrowRegionsWeighted");
   dip_sint32 *lptr;
   dip_sfloat *gptr;
   dip_sfloat *dptr;
   dip_sfloat newdistance;
   dip_int ii, jj;
   dip_IntegerArray coords, ncoords;
   dip_Boolean skip;
   dip_int ndims = imsz->size;
   dip_int nneighbours = NLcoords->size;
   void *vptr;
   DIP_INDEX_TO_COORDINATE_DECL( ix );
   DIP_FNR_INITIALISE;

   /* Walk over the entire image & put all the foreground border pixels on the queue
    * Also set the distance image to all DIP_MAX_SFLOAT */
   DIPXJ( dip_IntegerArrayNew( &coords, ndims, 0, rg ));
   for (;;) {
      coords->array[0] = 0;
      lptr = labels;
      gptr = grey;
      dptr = distance;
      for (ii=1; ii<ndims; ii++) {
         lptr += labstride->array[ii] * coords->array[ii];
         gptr += greystride->array[ii] * coords->array[ii];
         dptr += diststride->array[ii] * coords->array[ii];
      }
      for (jj=0; jj<imsz->array[0]; jj++) {
         DIPTS (*gptr < 0, "Negative grey-values illegal: would produce negative distances.");
         if (*lptr != 0 && dip__PixelHasBackgroundNeighbour( lptr, NLcoords, NLindices_lab, coords, imsz )) {
            DIPTS( PQT_Push( pqueue, (void*)dptr, 0 ), "Failed to push data onto the queue." );
#if DEBUG
            printf( "Pushing pixel at coords (%d,%d,%d)\n", (int)coords->array[0],
                                                            (int)coords->array[1],
                                                            (int)coords->array[2] );
            printf( "   label = %d\n", (int)*lptr );
#endif            
            *dptr = 0;
         }
         else {
            *dptr = DIP_MAX_SFLOAT;
         }
         lptr += labstride->array[0];
         gptr += greystride->array[0];
         dptr += diststride->array[0];
         coords->array[0]++;
      }
      for (ii=1; ii<ndims; ii++) {
         coords->array[ii]++;
         if (coords->array[ii] < imsz->array[ii]) {
            break;
         }
         coords->array[ii] = 0;
      }
      if (ii == ndims) {
         break;
      }
   }
#if DEBUG
   printf( "dip__GrowRegionsWeighted: Put %d pixels on the queue\n", PQT_NElements( pqueue ));
   fflush( stdout );
#endif

   /* Start processing pixels */
   DIPXJ( dip_IntegerArrayNew( &ncoords, ndims, 0, rg ));
   DIP_INDEX_TO_COORDINATE_INIT( imsz, diststride, ix, rg );
   for (;;) {

      /* Check ending condition */
      if (PQT_IsEmpty( pqueue )) {
#if DEBUG
         printf( "dip__GrowRegionsWeighted: Finished because the queue is empty\n" );
         fflush( stdout );
#endif
         break;
      }

      /* Get a pixel from the queue */
      DIPTS( PQT_Pop( pqueue, &vptr, 0 ), "Failed to pop data from the queue." );
      dptr = (dip_sfloat*)vptr;
      if (*dptr >= 0) {

         /* Compute the distance to unprocessed neighbors. Update the neighbors by
          * propagating the distance and label, and add them to the queue. */
         DIP_INDEX_TO_COORDINATE( dptr-distance, coords, diststride, ix );
         lptr = labels;
         gptr = grey;
         for (ii=0; ii<ndims; ii++) {
            lptr += labstride->array[ii] * coords->array[ii];
            gptr += greystride->array[ii] * coords->array[ii];
         }
         for (jj=0; jj<nneighbours; jj++) {
            skip = DIP_FALSE;
            for (ii=0; ii<ndims; ii++) {
               ncoords->array[ii] = coords->array[ii] + NLcoords->array[ii][jj];
               if ((ncoords->array[ii] < 0) ||
                   (ncoords->array[ii] > imsz->array[ii]-1)) {
                     skip = DIP_TRUE;
                     break;
               }
            }
            if (!skip) {
               if (*(dptr + NLindices_dist->array[jj]) >= 0) {
                  newdistance = *dptr + *(gptr + NLindices_grey->array[jj]) * NLdistances->array[jj];
                  if (*(dptr + NLindices_dist->array[jj]) > newdistance) {
                     *(dptr + NLindices_dist->array[jj]) = newdistance;
                     *(lptr + NLindices_lab->array[jj]) = *lptr;
                     DIPTS( PQT_Push( pqueue, (void*)(dptr+NLindices_dist->array[jj]),
                                 newdistance ), "Failed to push data onto the queue." );
                  }
               }
            }
         }

         /* Mark this pixels done */
         *dptr = -*dptr;

      }
      /* else it's been processed - it was popped before with a lower grey-value */

   }

dip_error:
   DIP_FNR_EXIT;
}


/*
 * The simplified version of the DIPlib function dip_GrowRegionsWeighted()
 */
static dip_Error GrowRegionsWeighted
(
   dip_Image lab,
   dip_Image grey
)
{
   DIP_FNR_DECLARE("GrowRegionsWeighted");
   dip_int ii, ndims, blocksize;
   dip_IntegerArray imsz, labstride, greystride, diststride;
   dip_IntegerArray NLindices_lab, NLindices_grey, NLindices_dist;
   dip_CoordinateArray NLcoords;
   dip_FloatArray NLdistances, pixelsize;
   dip_DataType dt;
   dip_Image distance;
   dip_ImageArray imar;
   dip_VoidPointerArray imdp;
   PQT_priorityqueue pqueue = 0;
   DIP_FNR_INITIALISE;

   /* Check inputs */
   DIPXJ( dip_IsScalar( lab, 0 ));
   DIPXJ( dip_ImageGetDataType( lab, &dt ));
   DIPTS( dt != DIP_DT_SINT32, "Image 'lab' must be 'sint32'.");
   DIPXJ( dip_ImageGetDimensions( lab, &imsz, rg ));
   ndims = imsz->size;
   DIPTS( ndims < 1, DIP_E_DIMENSIONALITY_NOT_SUPPORTED );
   DIPXJ( dip_ImagesCompareTwo( lab, grey, DIP_CPIM_SIZE_MATCH | DIP_CPIM_TYPES_MATCH, 0 ));
   DIPXJ( dip_ImageGetDataType( grey, &dt ));
   DIPTS( dt != DIP_DT_SFLOAT, "Image 'grey' must be 'sfloat'.");

   /* Prepare the list of neighbours */
   DIPXJ (dip_FloatArrayNew (&pixelsize, ndims, 1, rg));
   DIPXJ (dip_NeighbourListMakeChamfer (pixelsize, 2, &NLcoords, &NLdistances, rg));

   /* distance image */
   DIPXJ( dip_ImageNew( &distance, rg ));
   DIPXJ( dip_ImageCopyProperties( grey, distance ));
   DIPXJ( dip_ImageSetDataType( distance, DIP_DT_SFLOAT )); /* not necessary, because grey is sfloat too */
   DIPXJ( dip_ImageForge( distance ));

   /* Prepare the lists of indices to neighbours */
   DIPXJ( dip_ImageGetStride( lab, &labstride, rg ));
   DIPXJ( dip_NeighbourListToIndices( NLcoords, labstride, &NLindices_lab, rg ));
   DIPXJ( dip_ImageGetStride( grey, &greystride, rg ));
   DIPXJ( dip_NeighbourListToIndices( NLcoords, greystride, &NLindices_grey, rg ));
   DIPXJ( dip_ImageGetStride( distance, &diststride, rg ));
   DIPXJ( dip_NeighbourListToIndices( NLcoords, diststride, &NLindices_dist, rg ));

   /* Prepare the priority queue */
   blocksize = imsz->array[0];
   for (ii=1; ii<ndims; ii++) {
      blocksize *= imsz->array[ii];
   }
   blocksize = dipm_Sqrt( blocksize );
   DIPTS( PQT_New( &pqueue, blocksize ), "Failed to allocate priority queue." );

   /* Get data pointers */
   DIPXJ( dip_ImageArrayNew ( &imar, 3, rg ));
   imar->array[ 0 ] = lab;
   imar->array[ 1 ] = grey;
   imar->array[ 2 ] = distance;
   DIPXJ( dip_ImageGetData( imar, &imdp, 0, 0, 0, 0, 0, rg ));

   /* Do the thing */
   DIPXJ( dip__GrowRegionsWeighted ( imdp->array[ 0 ], imdp->array[ 1 ], imdp->array[ 2 ],
                       imsz, pqueue, labstride, greystride, diststride,
                       NLindices_lab, NLindices_grey, NLindices_dist, NLcoords, NLdistances ));

dip_error:
   if (pqueue) {
      PQT_Free( pqueue );
   }
   DIP_FNR_EXIT;
}


/*
 * Entry point
 */
int main (int argc, char *argv[]) {

   DIP_FN_DECLARE("main");
   dip_Resources rg = 0;
   dip_Image seed, grey;
   dip_IntegerArray dims, coord;
   dip_FloatArray fparam;
   clock_t begintime;
   dip_Random random;
   dip_float p;
   dip_int imsz, nseeds, ii;

   /* Get input parameters */
   if (argc>=2) {
      imsz = atoi (argv[1]);
      if (imsz<10) {
         imsz = 10;
      }
   }
   else {
      imsz = 1024;
   }

   /* Initialize DIPlib & allocate Resources structure*/
   DIPXJ( dip_Initialise( ));
   DIPXJ( dip_ResourcesNew( &rg, 0 ));
   random.highprecision = DIP_FALSE;
   DIPXJ( dip_RandomSeed( &random, 12345 ));

   /* Create grey-value image */
   DIPXJ( dip_ImageNew( &grey, rg ));
   DIPXJ( dip_ImageSetType( grey, DIP_IMTP_SCALAR ));
   DIPXJ( dip_ImageSetDataType( grey, DIP_DT_SFLOAT ));
   DIPXJ( dip_IntegerArrayNew( &dims, 3, imsz, rg ));
   DIPXJ( dip_ImageSetDimensions( grey, dims ));
   DIPXJ( dip_ImageForge( grey ));
   DIPXJ( dip_Clear( grey )); /* sets it all to 0 */
   DIPXJ( dip_GaussianNoise( grey, grey, 100, &random ));
   DIPXJ( dip_FloatArrayNew( &fparam, 3, 10, rg ));
   DIPXJ( dip_Gauss ( grey, grey, 0, 0, fparam, 0, 0 ));
   DIPXJ( dip_GaussianNoise( grey, grey, 1, &random ));
   DIPXJ( dip_GetMaximumAndMinimum ( grey, 0, 0, &p ));
   DIPXJ( dip_AddFloat ( grey, grey, (1e-6)-p ));

   /* Create label image */
   DIPXJ( dip_ImageNew( &seed, rg ));
   DIPXJ( dip_ImageCopyProperties( grey, seed ));
   DIPXJ( dip_ImageSetDataType( seed, DIP_DT_SINT32 ));
   DIPXJ( dip_ImageForge( seed ));
   DIPXJ( dip_Clear( seed )); /* sets it all to 0 */
   p = round(imsz/40);
   if (p<1) p=1;
   nseeds = (dip_int)(p*p*p); /* this is how many seeds I want */
#if DEBUG
   printf( "Number of pixels in image: %d\n", (int)imsz );
   printf( "Number of seed pixels: %d\n", (int)nseeds );
   fflush( stdout );
#else
   printf( "%d, %d, ", (int)imsz, (int)nseeds);
#endif
   DIPXJ( dip_IntegerArrayNew( &coord, 3, 0, rg ));
   for (ii=1; ii<=nseeds; ii++) {
      DIPXJ( dip_UniformRandomVariable ( &random, 0, dims->array[0], &p ));
      coord->array[0] = (dip_int)p;
      DIPXJ( dip_UniformRandomVariable ( &random, 0, dims->array[1], &p ));
      coord->array[1] = (dip_int)p;
      DIPXJ( dip_UniformRandomVariable ( &random, 0, dims->array[2], &p ));
      coord->array[2] = (dip_int)p;
      DIPXJ( dip_SetInteger ( seed, ii, coord, DIP_FALSE ));
   }

   /* Time the region growing code */
   begintime = clock ();
   DIPXJ( GrowRegionsWeighted( seed, grey ));
#if DEBUG
   printf( "Elapsed time: %g s.\n", (clock ()-begintime)/(double)CLOCKS_PER_SEC );
#else
   printf( "%f\n", (clock ()-begintime)/(double)CLOCKS_PER_SEC );
#endif

dip_error:

   /* Return all allocated memory to the system */
   DIPXC( dip_ResourcesFree (&rg) );

   /* Let libraries clean up after themselves */
   DIPXC( dip_Exit () );

   /* Report errors */
   return dip_ErrorWrite (error, 0, 0, stderr);
}
