/*
 * Testing the various priority queues
 *
 * Pushing NITEMS random numbers [0,50] on the queue.
 * Printing the memory footprint of the executable.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "pqt.h"

#define DEBUG 0

#define INIT_RANGE 50

void printusage (void) {
   printf ("Usage: pqt <queue size>\n");
}

int main (int argc, const char* argv[]) {

   PQT_priorityqueue pq;
   int nitems, ii;
   float item;

   if (argc!=2) {
      printusage ();
      return 1;
   }
   nitems = atoi (argv[1]);
   if (nitems <= 1) {
      printusage ();
      return 1;
   }

/*** Initialise ***/

#if DEBUG
   printf ("Creating object.\n");
#endif

   if (PQT_New (&pq, 0)) {
      printf ("Failed to create priority queue object!\n");
      return 1;
   }

   srand (1);

/*** Adding nitems items to queue ***/

#if DEBUG
   printf ("Pushing data onto queue.\n");
#endif

   for (ii=0; ii<nitems; ii++) {
      item = INIT_RANGE * ((float)rand ()) / RAND_MAX;
      if (PQT_Push (pq, 0, item)) {
         printf ("Failed to push item onto queue!\n");
         return 1;
      }
   }

#if DEBUG
   printf ("Using the queue a little bit.\n");
#endif

   for (ii=0; ii<100; ii++) {
      if (PQT_Pop (pq, 0, &item)) {
         printf ("Failed to pop item from queue!\n");
         return 1;
      }
      item = INIT_RANGE; /* if we always push max value here, we're sure it's not going to be < min */
      if (PQT_Push (pq, 0, item)) {
         printf ("Failed to push item onto queue!\n");
         return 1;
      }
   }

   printf ("%d,", PQT_NElements(pq));
   fflush (stdout);
   system ("ps --no-headers --format \"rss vsz\" -C pqt_memory");

/* Meaning of PS format options:
 *
rss        RSS      resident set size, the non-swapped physical memory that a task has
                    used (in kiloBytes). (alias rssize, rsz).

size       SZ       approximate amount of swap space that would be required if the process
                    were to dirty all writable pages and then be swapped out. This number
                    is very rough!

sz         SZ       size in physical pages of the core image of the process. This includes
                    text, data, and stack space. Device mappings are currently excluded;
                    this is subject to change. See vsz and rss.

vsz        VSZ      virtual memory size of the process in KiB (1024-byte units). Device
                    mappings are currently excluded; this is subject to change.
                    (alias vsize).
 *
 * This means that VSZ = 4*SZ (4Kb pages on this computer), and RSS is the portion of VSZ
 * that has actually been used. The SZ from "size" option is slightly smaller than VSZ.
 */

/*** Clean up ***/

#if DEBUG
   printf ("Cleaning up.\n");
#endif

   PQT_Free (pq);
   return 0;
}
