/*
 * Testing the various priority queues
 *
 * Pushing NITEMS random numbers [0,50] on the queue.
 * Timing how long it takes to do NITERATIONS pop/push cycles.
 * The value pushed is the value popped + some random value [0,2].
 * I'm hoping this sequence mimics an image processing routine
 * like the watershed or the grey-weighted distance transform.
 *
 * (c)2008-2009, Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * GNU GPL v.3.0 -- see gpl-3.0.txt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "pqt.h"

#define DEBUG 0

void printusage (void) {
   printf ("Usage: pqt <queue size> <iteration count> <init range> <increment range> <min increment>\n");
}

int main (int argc, const char* argv[]) {

   PQT_priorityqueue pq;
   int nitems, niterations, ii;
   float item;
   clock_t begintime;
#if DEBUG
   float previtem;
#endif
   float init_range = 50, increment_range = 2, min_increment = 0;

   if (argc!=3 && argc!=6) {
      printusage ();
      return 1;
   }
   nitems = atoi (argv[1]);
   if (nitems <= 1) {
      printusage ();
      return 1;
   }
   niterations = atoi (argv[2]);
   if (niterations < 0) {
      printusage ();
      return 1;
   }
   if (argc>3) {
      init_range      = atof (argv[3]);
      increment_range = atof (argv[4]);
      min_increment   = atof (argv[5]);
      if (init_range <= 0 || increment_range <= 0 || min_increment < 0) {
         printusage ();
         return 1;
      }
   }

/*** Initialise ***/

#if DEBUG
   printf ("Creating object.\n");
#endif

   if (PQT_New (&pq, 0)) {
      printf ("Failed to create priority queue object!\n");
      return 1;
   }

   srand (1);

/*** Timing: adding nitems items to queue ***/

#if DEBUG
   printf ("Pushing data onto queue.\n");
#endif

   begintime = clock ();
   for (ii=0; ii<nitems; ii++) {
      item = init_range * ((float)rand ()) / RAND_MAX;
      if (PQT_Push (pq, 0, item)) {
         printf ("Failed to push item onto queue!\n");
         return 1;
      }
   }

#if !DEBUG

   printf ("%d, %g, ", PQT_NElements(pq),(clock ()-begintime)/(double)CLOCKS_PER_SEC);

#else

   printf ("Pushed %d items on the queue.\n", PQT_NElements(pq));
   printf ("Elapsed time: %g s.\n",(clock ()-begintime)/(double)CLOCKS_PER_SEC);
   printf ("Memory used (RSS, VSZ, in KiB): ");
   fflush (stdout);
   system ("ps --no-headers --format \"rss vsz\" -C pqt_speed");

#endif /* DEBUG */

/*** Timing: pushing and popping niterations items from queue ***/

#if DEBUG
   printf ("Pushing and popping data on queue.\n");
   previtem = -1000; /* smaller than anything we've pushed onto the queue! */
#endif

   begintime = clock ();
   for (ii=0; ii<niterations; ii++) {
      if (PQT_Pop (pq, 0, &item)) {
         printf ("Failed to pop item from queue!\n");
         return 1;
      }
#if DEBUG==2
      printf ("%g ", item);
#endif
#if DEBUG
      if (item<previtem) {
         printf ("Popped item out of order!\n");
         return 1;
      }
      previtem = item;
#endif
      item += ( increment_range * ((float)rand ()) / RAND_MAX ) + min_increment;
      if (PQT_Push (pq, 0, item)) {
         printf ("Failed to push item onto queue!\n");
         return 1;
      }
#if DEBUG
      if (item<previtem) {
         previtem = item;
      }
#endif
   }
#if DEBUG==2
   printf ("\n");
#endif

#if !DEBUG

   printf ("%g, ",(clock ()-begintime)/(double)CLOCKS_PER_SEC);

#else

   printf ("There are currently %d items on the queue.\n", PQT_NElements(pq));
   printf ("Elapsed time: %g s.\n",(clock ()-begintime)/(double)CLOCKS_PER_SEC);
   printf ("Memory used (RSS, VSZ, in KiB): ");
   fflush (stdout);
   system ("ps --no-headers --format \"rss vsz\" -C pqt_speed");

#endif /* DEBUG */

/*** Timing: popping nitems items from queue ***/

#if DEBUG
   printf ("Popping data from queue.\n");
#endif

   begintime = clock ();
   for (ii=0; ii<nitems; ii++) {
      if (PQT_IsEmpty (pq)) {
         break;
      }
      if (PQT_Pop (pq, 0, &item)) {
         printf ("Failed to pop item from queue!\n");
         return 1;
      }
#if DEBUG==2
      printf ("%g ", item);
#endif
#if DEBUG
      if (item<previtem) {
         printf ("Popped item out of order!\n");
         return 1;
      }
      previtem = item;
#endif
   }
#if DEBUG==2
   printf ("\n");
#endif
#if DEBUG
   if (ii != nitems) {
      printf ("Popped %d items, instead of %d!\n", ii, nitems);
   }
   if (!PQT_IsEmpty (pq)) {
      printf ("The queue is not empty!\n");
   }
#endif

#if !DEBUG

   printf ("%g\n",(clock ()-begintime)/(double)CLOCKS_PER_SEC);

#else

   printf ("Elapsed time: %g s.\n",(clock ()-begintime)/(double)CLOCKS_PER_SEC);

#endif /* DEBUG */

/*** Clean up ***/

   PQT_Free (pq);
   return 0;
}
